using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Xml.Linq;
using EvaPosSCOSrv;
using Serilog;
using static System.Runtime.InteropServices.JavaScript.JSType;

namespace gatewaySCO.POSBC;

// Maneja conexión al POSBC.
public class ClienteServidorPOSBC(string ip, int pto)
{
    ILogger log = Log.ForContext<ClienteServidorPOSBC>();
    int _pto = pto;
    string _ip = ip;
    Socket _socket = null;
    public bool ConexionActiva { get; private set; } = false;
    int _contadorMensajesEnviados = 0;
    int _contadorMensajesRecibidos = 0;

    public void AbreConexion()
    {
        _socket = new(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
        // Conectarse al servidor
        try
        {
            _socket.Connect(new IPEndPoint(IPAddress.Parse(_ip), _pto));
            ConexionActiva = true;
            Log.Information("Conectado a POSBC {ip}:{pto} - {conectado}", _ip, _pto, _socket.Connected);
        }
        catch (SocketException ex)
        {
            Log.Error("Excepción socket conexión a POSBC {ip}:{pto} - código error socket {codigo} - {e}", _ip, ex.SocketErrorCode, _pto, ex);
        }
        catch (Exception e)
        {
            Log.Error("Excepción en conexión a POSBC {ip}:{pto} - {e}", _ip, _pto, e);
        }
    }

    public void CierraConexion()
    {
        // Cerrar el socket
        ConexionActiva = false;
        try
        {
            _socket.Shutdown(SocketShutdown.Both);
            _socket.Close();
        }
        catch (Exception e)
        {
            Log.Warning("Excepción cerrando conexión a POSBC {ip}:{pto} - {e}", _ip, _pto, e);
        }
        Log.Information("Conexión cerrada a POSBC {ip}:{pto} - {conectado}", _ip, _pto, _socket.Connected);
    }

    // Convierte el string del parámetro a un arreglo de bytes,
    // anteponiendo 4 bytes con la longitud del mensaje. 
    public static byte[] ConvertirEnBytes(string mensaje)
    {
        // Codifica longitud del mensaje en un entero sin signo en los 4 primeros bytes.
        var longitud = Convert.ToUInt32(mensaje.Length);
        byte[] bytesConLongMensaje = BitConverter.GetBytes(longitud);
        // Bytes mas significativos deben ir primero, usa 'big-endian'.
        if (BitConverter.IsLittleEndian)
            Array.Reverse(bytesConLongMensaje);
        // Codifica en bytes texto del mensaje.
        byte[] bytesConMensaje = Encoding.UTF8.GetBytes(mensaje);
        // Copia los 2 arreglos de bytes en un arreglo unificado.
        byte[] bytes = new byte[bytesConLongMensaje.Length + bytesConMensaje.Length];
        Buffer.BlockCopy(bytesConLongMensaje, 0, bytes, 0, bytesConLongMensaje.Length);
        Buffer.BlockCopy(bytesConMensaje, 0, bytes, bytesConLongMensaje.Length, bytesConMensaje.Length);
        return bytes;
    }

    // Envia mensaje, espera respuesta y la retorna como un arreglo de bytes.
    // El mensaje es un string, convertido a arreglo de bytes, y le antepone
    // 4 bytes con la longitud del mensaje.
    // TODO : validar que la respuesta sean mas de 1 mensaje!!!
    public byte[] EnviaRecibe(byte[] bytesMensaje)
    {
        // Remite mensaje.
        _socket.Send(bytesMensaje);
        _contadorMensajesEnviados++;
        Log.Information("Mensaje #{contador} para POSBIC {ip} - {msj}", _contadorMensajesEnviados, _ip, bytesMensaje.ToString);
        Log.Information("Esperando respuesta..");
        
        // Leer la respuesta del servidor
        byte[] buffer = new byte[1024]; // Tamaño del buffer inicial
        int totalBytesRead = 0;
        int bytesRead;
        byte[] bufferEntrada;

        // Usar un MemoryStream para manejar datos de longitud arbitraria
        using (var ms = new System.IO.MemoryStream(1024))
        {
            while ((bytesRead = _socket.Receive(buffer)) > 0)
            {
                ms.Write(buffer, 0, bytesRead);
                totalBytesRead += bytesRead;

                // Romper el ciclo si se ha leído todo el mensaje
                if (_socket.Available == 0) break;
            }

            // Obtener todos los datos recibidos como un arreglo de bytes
            bufferEntrada = ms.ToArray();
        }
        _contadorMensajesRecibidos++;
        string result = Encoding.UTF8.GetString(bufferEntrada);

        Log.Debug("Respuesta #{contador} de POSBC {ip} - {msj}", _contadorMensajesRecibidos, _ip, bufferEntrada.ToString);
        return bufferEntrada;
    }
}